#
# Copyright (C) 2014-2015  UAVCAN Development Team  <uavcan.org>
#
# This software is distributed under the terms of the MIT License.
#
# Author: Pavel Kirienko <pavel.kirienko@zubax.com>
#         Ben Dyer <ben_dyer@mac.com>
#

"""
This module implements a fully compliant DroneCAN DSDL parser.
Please read the specs at http://dronecan.org.
"""

from .parser import Parser, parse_namespaces, \
    Type, PrimitiveType, ArrayType, CompoundType, VoidType, \
    Attribute, Field, Constant

from .common import DsdlException, crc16_from_bytes

__all__ = ['Parser', 'parse_namespaces',
           'Type', 'PrimitiveType', 'ArrayType', 'CompoundType', 'VoidType',
           'Attribute', 'Field', 'Constant',
           'DsdlException', 'crc16_from_bytes']
